"""Define the API for the Python Factory."""

from fastapi import APIRouter

from .tags import TagEnum
from .v1.sys import api_v1_sys

api: APIRouter = APIRouter(prefix="/api")

### API v1 ###
# Prefix the API with /api/v1
api_v1: APIRouter = APIRouter(prefix="/v1")
api_v1.include_router(router=api_v1_sys)


### API v2 ###
# Prefix the API with /api/v2
api_v2: APIRouter = APIRouter(prefix="/v2")


### Include the API routers ###
api.include_router(router=api_v1)
api.include_router(router=api_v2)

__all__: list[str] = ["TagEnum", "api"]



"""Provides enums for the app module."""

from enum import StrEnum


class EnvironmentEnum(StrEnum):
    """Represents the environment."""

    DEVELOPMENT = "development"
    STAGING = "staging"
    PRODUCTION = "production"



"""OpenTelemetry Plugin Module."""

from .configs import OpenTelemetryConfig, OpenTelemetryMeterConfig, OpenTelemetryTracerConfig
from .exceptions import OpenTelemetryPluginBaseException, OpenTelemetryPluginConfigError
from .plugins import OpenTelemetryPlugin, depends_meter_provider, depends_otel_config, depends_tracer_provider

__all__: list[str] = [
    "OpenTelemetryConfig",
    "OpenTelemetryMeterConfig",
    "OpenTelemetryPlugin",
    "OpenTelemetryPluginBaseException",
    "OpenTelemetryPluginConfigError",
    "OpenTelemetryTracerConfig",
    "depends_meter_provider",
    "depends_otel_config",
    "depends_tracer_provider",
]



"""Python Factory Example."""

from fastapi_factory_utilities.example.app import AppBuilder


def main() -> None:
    """Main function."""
    AppBuilder().build_and_serve()


__all__: list[str] = ["main"]



"""Provides the exceptions for the Aiopika plugin."""

from typing import Any

from fastapi_factory_utilities.core.exceptions import FastAPIFactoryUtilitiesError


class AiopikaPluginBaseError(FastAPIFactoryUtilitiesError):
    """Base class for all exceptions raised by the Aiopika plugin."""

    def __init__(self, message: str, **kwargs: Any) -> None:
        """Initialize the Aiopika plugin base exception."""
        super().__init__(message, **kwargs)


class AiopikaPluginConfigError(AiopikaPluginBaseError):
    """Exception for the Aiopika plugin configuration."""


class AiopikaPluginConnectionNotProvidedError(AiopikaPluginBaseError):
    """Exception for the Aiopika plugin connection not provided."""


class AiopikaPluginExchangeNotDeclaredError(AiopikaPluginBaseError):
    """Exception for the Aiopika plugin exchange not declared."""


class AiopikaPluginQueueNotDeclaredError(AiopikaPluginBaseError):
    """Exception for the Aiopika plugin queue not declared."""



"""Provide Kratos Session and Identity classes."""

from enum import StrEnum
from typing import Annotated

from fastapi import Depends, HTTPException, Request

from fastapi_factory_utilities.core.services.kratos import (
    KratosOperationError,
    KratosService,
    KratosSessionInvalidError,
    KratosSessionObject,
    depends_kratos_service,
)


class KratosSessionAuthenticationErrors(StrEnum):
    """Kratos Session Authentication Errors."""

    MISSING_CREDENTIALS = "Missing Credentials"
    INVALID_CREDENTIALS = "Invalid Credentials"
    INTERNAL_SERVER_ERROR = "Internal Server Error"


class KratosSessionAuthentication:
    """Kratos Session class."""

    DEFAULT_COOKIE_NAME: str = "ory_kratos_session"

    def __init__(self, cookie_name: str = DEFAULT_COOKIE_NAME, raise_exception: bool = True) -> None:
        """Initialize the KratosSessionAuthentication class.

        Args:
            cookie_name (str): Name of the cookie to extract the session
            raise_exception (bool): Whether to raise an exception or return None
        """
        self._cookie_name: str = cookie_name
        self._raise_exception: bool = raise_exception

    def _extract_cookie(self, request: Request) -> str | None:
        """Extract the cookie from the request.

        Args:
            request (Request): FastAPI request object.

        Returns:
            str | None: Cookie value or None if not found.

        Raises:
            HTTPException: If the cookie is missing.
        """
        return request.cookies.get(self._cookie_name, None)

    async def __call__(
        self, request: Request, kratos_service: Annotated[KratosService, Depends(depends_kratos_service)]
    ) -> KratosSessionObject | KratosSessionAuthenticationErrors:
        """Extract the Kratos session from the request.

        Args:
            request (Request): FastAPI request object.
            kratos_service (KratosService): Kratos service object.

        Returns:
            KratosSessionObject | KratosSessionAuthenticationErrors: Kratos session object or error.

        Raises:
            HTTPException: If the session is invalid and raise_exception is True.
        """
        cookie: str | None = self._extract_cookie(request)
        if not cookie:
            if self._raise_exception:
                raise HTTPException(
                    status_code=401,
                    detail=KratosSessionAuthenticationErrors.MISSING_CREDENTIALS,
                )
            else:
                return KratosSessionAuthenticationErrors.MISSING_CREDENTIALS

        try:
            session: KratosSessionObject = await kratos_service.whoami(cookie_value=cookie)
        except KratosSessionInvalidError as e:
            if self._raise_exception:
                raise HTTPException(
                    status_code=401,
                    detail="Invalid Credentials",
                ) from e
            else:
                return KratosSessionAuthenticationErrors.INVALID_CREDENTIALS
        except KratosOperationError as e:
            if self._raise_exception:
                raise HTTPException(
                    status_code=500,
                    detail="Internal Server Error",
                ) from e
            else:
                return KratosSessionAuthenticationErrors.INTERNAL_SERVER_ERROR

        return session



"""Provide importlib functions."""

from importlib.resources import files
from pathlib import Path


def get_path_file_in_package(filename: str, package: str) -> Path:
    """Return Absolute Path of file in package.

    Args:
        filename (str): Filename to search
        package (str): Package name

    Returns:
        Traversable: File

    Raises:
        FileNotFoundError: If file not found
        ImportError: If package not found

    """
    try:
        path: Path = Path(str(files(package).joinpath(filename)))
    except FileNotFoundError as exception:
        raise FileNotFoundError(f"File {filename} not found in package {package}") from exception
    except ImportError as exception:
        raise ImportError(f"Package {package} not found") from exception
    return path



"""Package for system related API endpoints."""

from fastapi import APIRouter

from .health import api_v1_sys_health
from .readiness import api_v1_sys_readiness

api_v1_sys = APIRouter(prefix="/sys")
api_v1_sys.include_router(router=api_v1_sys_health)
api_v1_sys.include_router(router=api_v1_sys_readiness)

__all__: list[str] = ["api_v1_sys"]



"""API v1 sys readiness module.

Provide the Get readiness endpoint
"""

from http import HTTPStatus

from fastapi import APIRouter, Depends, Response
from pydantic import BaseModel

from fastapi_factory_utilities.core.services.status.enums import ReadinessStatusEnum
from fastapi_factory_utilities.core.services.status.services import (
    StatusService,
    depends_status_service,
)

api_v1_sys_readiness = APIRouter(prefix="/readiness")


class ReadinessResponseModel(BaseModel):
    """Readiness response schema."""

    status: ReadinessStatusEnum


@api_v1_sys_readiness.get(
    path="",
    tags=["sys"],
    response_model=ReadinessResponseModel,
    responses={
        HTTPStatus.OK.value: {
            "model": ReadinessResponseModel,
            "description": "Readiness status.",
        },
        HTTPStatus.INTERNAL_SERVER_ERROR.value: {
            "model": ReadinessResponseModel,
            "description": "Internal server error.",
        },
    },
)
def get_api_v1_sys_readiness(
    response: Response, status_service: StatusService = Depends(depends_status_service)
) -> ReadinessResponseModel:
    """Get the readiness of the system.

    Args:
        response (Response): The response object.
        status_service (StatusService): The status service

    Returns:
        ReadinessResponse: The readiness status.
    """
    status: ReadinessStatusEnum = status_service.get_status()["readiness"]

    match status:
        case ReadinessStatusEnum.READY:
            response.status_code = HTTPStatus.OK.value
        case ReadinessStatusEnum.NOT_READY:
            response.status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
    return ReadinessResponseModel(status=status)



"""Provides types for the books service."""

from typing import Any, cast

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class BookName(str):
    """Book name type."""

    MIN_LENGTH: int = 1
    MAX_LENGTH: int = 100

    def __new__(cls, value: str) -> "BookName":
        """Create a new instance of BookName."""
        return super().__new__(cls, cls.validate(value))

    @classmethod
    def __get_pydantic_core_schema__(cls, source: type[Any], handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        """Get the Pydantic core schema for the book name.

        Args:
            source (Type[Any]): Source type.
            handler (GetCoreSchemaHandler): Handler.

        Returns:
            core_schema.CoreSchema: Core schema.
        """
        del source, handler
        return cast(
            core_schema.CoreSchema,
            core_schema.no_info_after_validator_function(
                function=cls.validate,
                schema=core_schema.str_schema(min_length=cls.MIN_LENGTH, max_length=cls.MAX_LENGTH),
            ),
        )

    @classmethod
    def validate(cls, value: str) -> str:
        """Validate the book name.

        Args:
            value (str): Book name.

        Returns:
            str: Book name.
        """
        if not cls.MIN_LENGTH <= len(value) <= cls.MAX_LENGTH:
            raise ValueError(
                f"Expected a string with length between {cls.MIN_LENGTH}" + f" and {cls.MAX_LENGTH}, got {len(value)}"
            )

        return value



"""Provides the status types for the service."""

import datetime
from typing import ClassVar, NewType, TypedDict, cast

from pydantic import BaseModel, ConfigDict

from .enums import ComponentTypeEnum, HealthStatusEnum, ReadinessStatusEnum

ComponentInstanceKey = NewType("ComponentInstanceKey", str)


class ComponentInstanceType:
    """Component instance type.

    Attributes:
        component_type (ComponentTypeEnum): The component type.
        identifier (str | None): The identifier.
        key (ComponentInstanceKey): The key based on the component type and identifier.
    """

    def _generate_key(self) -> ComponentInstanceKey:
        """Generate the key identifier for the component instance.

        It is based on the component type and identifier.

        Returns:
            str: The key.

        """
        key: str = (
            f"{self._component_type.value}:{self._identifier}" if self._identifier else self._component_type.value
        )
        return cast(ComponentInstanceKey, key)

    def __init__(self, component_type: ComponentTypeEnum, identifier: str | None = None) -> None:
        """Initialize the component instance type.

        Args:
            component_type (ComponentTypeEnum): The component type.
            identifier (str, optional): The identifier. Defaults to None.

        """
        self._component_type: ComponentTypeEnum = component_type
        self._identifier: str | None = identifier
        self._key: ComponentInstanceKey = self._generate_key()

    @property
    def component_type(self) -> ComponentTypeEnum:
        """Get the component type.

        Returns:
            ComponentTypeEnum: The component type.

        """
        return self._component_type

    @property
    def identifier(self) -> str | None:
        """Get the identifier.

        Returns:
            str | None: The identifier.

        """
        return self._identifier

    @property
    def key(self) -> ComponentInstanceKey:
        """Get the key.

        Returns:
            str: The key.

        """
        return self._key


class Status(TypedDict):
    """Status type.

    Attributes:
        health (HealthStatusEnum): The health status.
        readiness (ReadinessStatusEnum): The readiness status.
    """

    health: HealthStatusEnum
    readiness: ReadinessStatusEnum


class StatusUpdateEvent(BaseModel):
    """Status update event."""

    # Pydantic config
    model_config: ClassVar[ConfigDict] = {"frozen": True}

    # Health status
    health_status: HealthStatusEnum
    previous_health_status: HealthStatusEnum

    # Readiness status
    readiness_status: ReadinessStatusEnum
    previous_readiness_status: ReadinessStatusEnum

    # Timestamp
    triggered_at: datetime.datetime


class ComponentInstanceStatusUpdateEvent(BaseModel):
    """Component instance status update event."""

    # Pydantic config
    model_config: ClassVar[ConfigDict] = {
        "frozen": True,
        "arbitrary_types_allowed": True,  # Needed for the ComponentInstanceType
    }

    # Component instance
    component_instance: ComponentInstanceType

    # Health status
    health_status: HealthStatusEnum

    # Readiness status
    readiness_status: ReadinessStatusEnum

    # Timestamp
    triggered_at: datetime.datetime



"""Provides the configuration for the Aiopika plugin."""

from typing import Annotated, Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, UrlConstraints
from pydantic_core import Url, ValidationError

from fastapi_factory_utilities.core.utils.importlib import get_path_file_in_package
from fastapi_factory_utilities.core.utils.yaml_reader import (
    UnableToReadYamlFileError,
    YamlFileReader,
)

from .exceptions import AiopikaPluginConfigError


class AiopikaConfig(BaseModel):
    """Provides the configuration model for the Aiopika plugin.

    https://docs.aio-pika.com/#aio-pika-connect-robust-function-and-aio-pika-robustconnection-class-specific

    Possible query parameters for the AMQP URL:
        name (str url encoded) - A string that will be visible in the RabbitMQ management console
        and in the server logs, convenient for diagnostics.
        cafile (str) - Path to Certificate Authority file
        capath (str) - Path to Certificate Authority directory
        cadata (str url encoded) - URL encoded CA certificate content
        keyfile (str) - Path to client ssl private key file
        certfile (str) - Path to client ssl certificate file
        no_verify_ssl - No verify server SSL certificates. 0 by default and means False other value means True.
        heartbeat (int-like) - interval in seconds between AMQP heartbeat packets. 0 disables this feature.
        reconnect_interval (float-like) - is the period in seconds, not more often than the attempts
        to re-establish the connection will take place.
        fail_fast (true/yes/y/enable/on/enabled/1 means True, otherwise False) - special behavior
        for the start connection attempt, if it fails, all other attempts stops
        and an exception will be thrown at the connection stage. Enabled by default, if you are sure you need
        to disable this feature, be ensures for the passed URL is really working.
        Otherwise, your program will go into endless reconnection attempts that can not be successed.

    """

    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    amqp_url: Annotated[Url, UrlConstraints(allowed_schemes=["amqp", "amqps"])] = Field(description="The AMQP URL.")


def build_config_from_package(package_name: str) -> AiopikaConfig:
    """Build the configuration from the package.

    Args:
        package_name (str): The package name.

    Returns:
        AiopikaConfig: The Aiopika configuration.

    Raises:
        AiopikaPluginConfigError: If the configuration cannot be read or created or the configuration is invalid.
    """
    try:
        yaml_file_content: dict[str, Any] = YamlFileReader(
            file_path=get_path_file_in_package(
                filename="application.yaml",
                package=package_name,
            ),
            yaml_base_key="aiopika",
            use_environment_injection=True,
        ).read()
    except (FileNotFoundError, ImportError, UnableToReadYamlFileError) as exception:
        raise AiopikaPluginConfigError(
            message="Unable to read the application configuration file for the Aiopika plugin in the package.",
            package_name=package_name,
        ) from exception

    # Create the application configuration model
    config: AiopikaConfig
    try:
        config = AiopikaConfig(**yaml_file_content)
    except ValidationError as exception:
        raise AiopikaPluginConfigError(
            message="Unable to create the application configuration model for the Aiopika plugin in the package.",
            package_name=package_name,
            validation_errors=exception.errors(),
        ) from exception

    return config



"""Provides base document class for ODM plugins."""

import datetime
from typing import Annotated
from uuid import UUID, uuid4

from beanie import Document, Indexed  # pyright: ignore[reportUnknownVariableType]
from pydantic import Field
from pymongo import DESCENDING


class BaseDocument(Document):
    """Base document class."""

    # To be agnostic of MongoDN, we use UUID as the document ID.
    id: UUID = Field(  # type: ignore
        default_factory=uuid4, description="The document ID."
    )

    revision_id: UUID | None = Field(default=None, exclude=False)
    created_at: Annotated[datetime.datetime, Indexed(index_type=DESCENDING)] = Field(  # pyright: ignore
        default_factory=lambda: datetime.datetime.now(tz=datetime.UTC), description="Creation timestamp."
    )

    updated_at: Annotated[datetime.datetime, Indexed(index_type=DESCENDING)] = Field(  # pyright: ignore
        default_factory=lambda: datetime.datetime.now(tz=datetime.UTC), description="Last update timestamp."
    )

    class Settings:
        """Meta class for BaseDocument."""

        use_revision = True



"""Provide the configuration for the app server."""

from typing import Any, ClassVar, Generic, TypeVar, get_args

from fastapi import Request
from pydantic import BaseModel, ConfigDict, Field, HttpUrl

from fastapi_factory_utilities.core.app.exceptions import ConfigBuilderError
from fastapi_factory_utilities.core.utils.configs import (
    UnableToReadConfigFileError,
    ValueErrorConfigError,
    build_config_from_file_in_package,
)
from fastapi_factory_utilities.core.utils.log import LoggingConfig

from .enums import EnvironmentEnum


def default_allow_all() -> list[str]:
    """Default allow all."""
    return ["*"]


class CorsConfig(BaseModel):
    """CORS configuration."""

    allow_origins: list[str] = Field(default_factory=default_allow_all, description="Allowed origins")
    allow_credentials: bool = Field(default=True, description="Allow credentials")
    allow_methods: list[str] = Field(default_factory=default_allow_all, description="Allowed methods")
    allow_headers: list[str] = Field(default_factory=default_allow_all, description="Allowed headers")
    expose_headers: list[str] = Field(default_factory=list, description="Exposed headers")
    max_age: int = Field(default=600, description="Max age")


class ServerConfig(BaseModel):
    """Server configuration."""

    # Pydantic configuration
    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    # Server configuration mainly used by uvicorn
    host: str = Field(default="0.0.0.0", description="Server host")
    port: int = Field(default=8000, description="Server port")
    workers: int = Field(default=1, description="Number of workers")


class DevelopmentConfig(BaseModel):
    """Development configuration."""

    # Pydantic configuration
    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    # Development configuration
    debug: bool = Field(default=False, description="Debug mode")
    reload: bool = Field(default=False, description="Reload mode")


class BaseApplicationConfig(BaseModel):
    """Application configuration abstract class."""

    # Pydantic configuration
    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="forbid")

    # Application configuration
    # (mainly used for monitoring and information reporting)
    service_namespace: str = Field(description="Service namespace")
    environment: EnvironmentEnum = Field(description="Deployed environment")
    service_name: str = Field(description="Service name")
    description: str = Field(description="Service description")
    version: str = Field(description="Service version")
    # Root path for the application
    root_path: str = Field(default="", description="Root path")


class HttpServiceDependencyConfig(BaseModel):
    """Http service dependency config."""

    url: HttpUrl


class DependencyConfig(BaseModel):
    """Dependency config."""

    kratos: HttpServiceDependencyConfig | None = Field(default=None, description="Kratos dependency config")
    hydra_admin: HttpServiceDependencyConfig | None = Field(default=None, description="Hydra admin dependency config")
    hydra_public: HttpServiceDependencyConfig | None = Field(default=None, description="Hydra public dependency config")


class RootConfig(BaseModel):
    """Root configuration."""

    # Pydantic configuration
    # extra = Extra.ignore, to be able to add extra categories for your application purposes
    model_config: ClassVar[ConfigDict] = ConfigDict(frozen=True, extra="ignore")

    # Root configuration with all sub configurations
    application: BaseApplicationConfig = Field(description="Application configuration")
    server: ServerConfig = Field(description="Server configuration", default_factory=ServerConfig)
    cors: CorsConfig = Field(description="CORS configuration", default_factory=CorsConfig)
    development: DevelopmentConfig = Field(description="Development configuration", default_factory=DevelopmentConfig)
    logging: list[LoggingConfig] = Field(description="Logging configuration", default_factory=list)
    dependencies: DependencyConfig = Field(description="Dependencies configuration", default_factory=DependencyConfig)


GenericConfig = TypeVar("GenericConfig", bound=BaseModel)


class GenericConfigBuilder(Generic[GenericConfig]):
    """Application configuration builder.

    This class is used to build the application configuration from a YAML file.
    It can be used to build any configuration model
    """

    DEFAULT_FILENAME: str = "application.yaml"
    DEFAULT_YAML_BASE_KEY: str | None = None

    def __init__(
        self,
        package_name: str,
        config_class: type[GenericConfig] | None = None,
        filename: str = DEFAULT_FILENAME,
        yaml_base_key: str | None = DEFAULT_YAML_BASE_KEY,
    ) -> None:
        """Instantiate the builder.

        Args:
            package_name (str): The package name.
            config_class (Type[AppConfigAbstract]): The configuration class.
            filename (str, optional): The filename. Defaults to DEFAULT_FILENAME.
            yaml_base_key (str, optional): The YAML base key. Defaults to DEFAULT_YAML_BASE_KEY.

        TODO: prevent the double definition of config_class and through the generic type
        """
        self.package_name: str = package_name
        generic_args: tuple[Any, ...] = get_args(self.__orig_bases__[0])  # type: ignore

        self.config_class: type[GenericConfig] = config_class if config_class is not None else generic_args[0]
        self.filename: str = filename
        self.yaml_base_key: str | None = yaml_base_key

    def build(self) -> GenericConfig:
        """Build the configuration.

        Returns:
            GenericConfig: The configuration.

        Raises:
            ApplicationConfigFactoryException: Any error occurred
        """
        try:
            config: GenericConfig = build_config_from_file_in_package(
                package_name=self.package_name,
                config_class=self.config_class,
                filename=self.filename,
                yaml_base_key=self.yaml_base_key,
            )
        except UnableToReadConfigFileError as exception:
            raise ConfigBuilderError(
                message="Unable to read the application configuration file.",
                config_class=self.config_class,
                package=self.package_name,
                filename=self.filename,
            ) from exception
        except ValueErrorConfigError as exception:
            raise ConfigBuilderError(
                message="Value error when creating the configuration object.",
                config_class=self.config_class,
                package=self.package_name,
                filename=self.filename,
            ) from exception
        except Exception as exception:
            raise ConfigBuilderError(
                message="An error occurred while building the application configuration.",
                config_class=self.config_class,
                package=self.package_name,
                filename=self.filename,
            ) from exception

        return config


def depends_dependency_config(request: Request) -> DependencyConfig:
    """Get the dependency config."""
    return request.app.state.config.dependencies



"""Provides the ApplicationAbstract class."""

from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any, ClassVar

from beanie import Document
from fastapi import FastAPI
from structlog.stdlib import BoundLogger, get_logger

from fastapi_factory_utilities.core.api import api
from fastapi_factory_utilities.core.app.config import RootConfig
from fastapi_factory_utilities.core.app.fastapi_builder import FastAPIBuilder
from fastapi_factory_utilities.core.plugins import PluginAbstract
from fastapi_factory_utilities.core.services.status.services import StatusService

_logger: BoundLogger = get_logger(__name__)


class ApplicationAbstract(ABC):
    """Application abstract class."""

    PACKAGE_NAME: ClassVar[str]

    CONFIG_CLASS: ClassVar[type[RootConfig]] = RootConfig

    # TODO: Find a way to remove this from here
    ODM_DOCUMENT_MODELS: ClassVar[list[type[Document]]]

    def __init__(
        self,
        root_config: RootConfig,
        plugins: list[PluginAbstract],
        fastapi_builder: FastAPIBuilder,
    ) -> None:
        """Instantiate the application."""
        self.config: RootConfig = root_config
        self.fastapi_builder: FastAPIBuilder = fastapi_builder
        # Add the API router to the FastAPI application
        self.fastapi_builder.add_api_router(router=api, without_resource_path=True)
        self.plugins: list[PluginAbstract] = plugins
        self._add_to_state: dict[str, Any] = {}

    def load_plugins(self) -> None:
        """Load the plugins."""
        for plugin in self.plugins:
            plugin.set_application(application=self)
            plugin.on_load()

    async def startup_plugins(self) -> None:
        """Startup the plugins."""
        for plugin in self.plugins:
            await plugin.on_startup()

    async def shutdown_plugins(self) -> None:
        """Shutdown the plugins."""
        for plugin in self.plugins:
            await plugin.on_shutdown()

    def setup(self) -> None:
        """Initialize the application."""
        # Initialize FastAPI application
        self._asgi_app: FastAPI = self.fastapi_builder.build(lifespan=self.fastapi_lifespan)
        # Status service
        self.status_service: StatusService = StatusService()
        self.add_to_state(key="status_service", value=self.status_service)
        self.add_to_state(key="config", value=self.config)
        self.add_to_state(key="application", value=self)
        self._apply_states_to_fastapi_app()
        # Configure the application
        self.configure()
        self._apply_states_to_fastapi_app()
        # Initialize PluginManager
        self.load_plugins()

    def _apply_states_to_fastapi_app(self) -> None:
        # Add manually added states to the FastAPI app
        for key, value in self._add_to_state.items():
            if hasattr(self._asgi_app.state, key):
                _logger.warn(f"Key {key} already exists in the state. Don't set it outside of the application.")
            setattr(self._asgi_app.state, key, value)
        self._add_to_state.clear()

    def add_to_state(self, key: str, value: Any) -> None:
        """Add a value to the FastAPI app state."""
        if key in self._add_to_state:
            raise ValueError(f"Key {key} already exists in the state.")
        self._add_to_state[key] = value

    @abstractmethod
    def configure(self) -> None:
        """Configure the application."""
        raise NotImplementedError

    @abstractmethod
    async def on_startup(self) -> None:
        """Startup the application."""
        raise NotImplementedError

    @abstractmethod
    async def on_shutdown(self) -> None:
        """Shutdown the application."""
        raise NotImplementedError

    @asynccontextmanager
    async def fastapi_lifespan(self, fastapi: FastAPI) -> AsyncGenerator[None, None]:  # pylint: disable=unused-argument
        """FastAPI lifespan context manager."""
        await self.startup_plugins()
        await self.on_startup()
        try:
            yield
        finally:
            await self.on_shutdown()
            await self.shutdown_plugins()

    def get_config(self) -> RootConfig:
        """Get the configuration."""
        return self.config

    def get_asgi_app(self) -> FastAPI:
        """Get the ASGI application."""
        return self._asgi_app

    def get_status_service(self) -> StatusService:
        """Get the status service."""
        return self.status_service

    async def __call__(self, scope: Any, receive: Any, send: Any) -> None:
        """Forward the call to the FastAPI app."""
        return await self._asgi_app.__call__(scope=scope, receive=receive, send=send)



"""Provides the configuration for the ODM plugin."""

from pydantic import BaseModel, ConfigDict


class ODMConfig(BaseModel):
    """Provides the configuration model for the ODM plugin."""

    model_config = ConfigDict(frozen=True, extra="forbid")

    uri: str

    database: str = "test"

    connection_timeout_ms: int = 4000



"""Provides the dependencies for the Aiopika plugin."""

from typing import cast

from aio_pika.abc import AbstractRobustConnection
from fastapi import Request

from .exceptions import AiopikaPluginBaseError

DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY: str = "aiopika_robust_connection"


def depends_aiopika_robust_connection(request: Request) -> AbstractRobustConnection:
    """Get the Aiopika robust connection."""
    robust_connection: AbstractRobustConnection | None = cast(
        AbstractRobustConnection | None, getattr(request.app.state, DEPENDS_AIOPIKA_ROBUST_CONNECTION_KEY, None)
    )
    if robust_connection is None:
        raise AiopikaPluginBaseError("Aiopika robust connection not found in the application state.")
    return robust_connection
